// 67d7842dbbe25473c3c32b93c0da8047785f30d78e8a024de1b57352245f9689
/*******************************************************************************
 * Copyright (c) 2007, 2016 Wind River Systems, Inc. and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 * Contributors:
 *     Wind River Systems - initial API and implementation
 *******************************************************************************/

/*******************************************************************************
 * Modified by Xilinx,Inc
 * Copyright 2022 Xilinx, Inc. All Rights Reserved. Modifications
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 *******************************************************************************/

package org.eclipse.tcf.internal.dap;

import org.eclipse.tcf.internal.dap.breakpoint.BreakpointInfo;
import org.eclipse.tcf.protocol.IChannel;
import org.eclipse.tcf.protocol.IToken;
import org.eclipse.tcf.protocol.Protocol;
import org.eclipse.tcf.services.IBreakpoints;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.function.Predicate;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Collectors;

/**
 * TCFBreakpointsModel class handles breakpoints for all active TCF launches.
 * It downloads initial set of breakpoint data when launch is activated,
 * listens for Eclipse breakpoint manager events and propagates breakpoint changes to TCF targets.
 */
public class TCFBreakpointsModel {

    private static final Logger LOGGER = Logger.getLogger(TCFBreakpointsModel.class.getSimpleName());

    public static final String
            CDATA_CLIENT_ID    = "ClientID",
            CDATA_TYPE         = "Type",
            CDATA_FILE         = "File",
            CDATA_MARKER       = "Marker";

    //    TODO:rm - private final IBreakpointManager bp_manager = DebugPlugin.getDefault().getBreakpointManager();
    private final HashMap<IChannel,Map<String,Object>> channels = new HashMap<IChannel,Map<String,Object>>();
//    TODO:rm - private final HashMap<String,IBreakpoint> id2bp = new HashMap<String,IBreakpoint>();
    private final HashMap<String, BreakpointInfo> id2bp = new HashMap<>();

    private abstract class BreakpointUpdate implements Runnable {

        private final BreakpointInfo breakpoint;
        private boolean removed;
        protected final boolean is_local;
        protected final String id;
        private final String file;
//        private final String type;

        IBreakpoints service;
        IBreakpoints.DoneCommand done;
        Map<String,Object> tcf_attrs;

        BreakpointUpdate(BreakpointInfo breakpoint, boolean removed) {
            this.breakpoint = breakpoint;
            this.removed = removed;
            is_local = !breakpoint.isForeign();
            file = breakpoint.getFile();
            id = breakpoint.getBreakpointId();
//            marker_type = breakpoint.getMarker().getType();
        }

        synchronized void exec() throws InterruptedException {
            assert !Protocol.isDispatchThread();
            if (id != null) {
                Protocol.invokeLater(this);
                wait();
            }
        }

        public void run() {
            if (disposed) return;
            if (removed) id2bp.remove(id);
            else id2bp.put(id, breakpoint);
            if (is_local) {
                for (final IChannel channel : channels.keySet()) {
                    tcf_attrs = toBreakpointAttributes(channel, breakpoint);
                    service = channel.getRemoteService(IBreakpoints.class);
                    if (!isSupported(channel, breakpoint)) continue;
                    done = new IBreakpoints.DoneCommand() {
                        public void doneCommand(IToken token, Exception error) {
                            if (error != null) channel.terminate(error);
                        }
                    };
                    update();
                }
            }
            Protocol.sync(new Runnable() {
                public void run() {
                    synchronized (BreakpointUpdate.this) {
                        BreakpointUpdate.this.notify();
                    }
                }
            });
        };

        abstract void update();
    }

    private void addBreakpoint(BreakpointInfo breakpoint) {
        try {
            new BreakpointUpdate(breakpoint, false) {
                @Override
                void update() {
                    service.add(tcf_attrs, done);
                }
            }.exec();
        }
        catch (Throwable x) {
            LOGGER.log(Level.SEVERE, "Unhandled exception in breakpoint listener", x);
        }
    }

    private void changeBreakpoint(BreakpointInfo breakpoint) {
        try {
            new BreakpointUpdate(breakpoint, false) {
                @Override
                void update() {
                    service.change(tcf_attrs, done);
                }
            }.exec();
        }
        catch (Throwable x) {
            LOGGER.log(Level.SEVERE, "Unhandled exception in breakpoint listener", x);
        }
    }

    private void removeBreakpoint(BreakpointInfo breakpoint) {
        try {
            new BreakpointUpdate(breakpoint, true) {
                @Override
                void update() {
                    service.remove(new String[]{ breakpoint.getBreakpointId() }, done);
                }
            }.exec();
        }
        catch (Throwable x) {
            LOGGER.log(Level.SEVERE, "Unhandled exception in breakpoint listener", x);
        }
    }

    private boolean disposed;

    public static final TCFBreakpointsModel INSTANCE = new TCFBreakpointsModel();
    public static TCFBreakpointsModel getBreakpointsModel() {
        return INSTANCE;
    }

    public void dispose() {
        channels.clear();
        disposed = true;
    }

    public boolean isSupported(IChannel channel, BreakpointInfo bp) {
        // TODO: implement per-channel breakpoint filtering
        return true;
    }

    /**
     * Get IBreakpoint for given TCF breakpoint ID.
     * The mapping works only for breakpoints that were sent to (or received from) a debug target.
     * It can be used to map target responses to IBreakpoint objects.
     * @param id - TCF breakpoint ID.
     * @return IBreakpoint object associated with the ID, or null.
     */
    public BreakpointInfo getBreakpoint(String id) {
        assert Protocol.isDispatchThread();
        return id2bp.get(id);
    }

    public void setBreakpoint(String id, BreakpointInfo bp) {
        assert Protocol.isDispatchThread();
        if (id2bp.containsKey(id)) {
            return;
        }
        id2bp.put(id, bp);
    }

/**
     * Check breakpoint ownership.
     * @param properties - breakpoint properties as reported by TCF Breakpoints service.
     * @return true if the breakpoint is owned by local instance of Eclipse.
     * Return false if the properties represent a breakpoint created by some other TCF client.
     */
    @SuppressWarnings("unchecked")
    public static boolean isLocal(Map<String,Object> properties) {
        if (properties == null) return false;
        Map<String,Object> client_data = (Map<String,Object>)properties.get(IBreakpoints.PROP_CLIENT_DATA);
        if (client_data == null) return false;
        String id = (String)client_data.get(TCFBreakpointsModel.CDATA_CLIENT_ID);
        return "tcf-dap-1".equals(id);
    }

/**
     * Check breakpoint marker ownership.
     * @param marker - breakpoint marker.
     * @return true if the marker is owned by local instance of Eclipse.
     * Return false if the marker represents a breakpoint created by some other TCF client.
     */
/*
    public static boolean isLocal(IMarker marker) {
        try {
            Map<String,Object> marker_attrs = marker.getAttributes();
            if (marker_attrs == null) return false;
            return !Boolean.FALSE.equals(marker_attrs.get(IBreakpoint.PERSISTED));
        }
        catch (CoreException e) {
            return false;
        }
    }
*/

    public void initialize(final IChannel channel, final Runnable done) {
        assert !disposed;
        assert Protocol.isDispatchThread();
        final IBreakpoints service = channel.getRemoteService(IBreakpoints.class);
        if (service == null) {
            Protocol.invokeLater(done);
            return;
        }
        service.getCapabilities(null, new IBreakpoints.DoneGetCapabilities() {
            public void doneGetCapabilities(IToken token, Exception error, Map<String, Object> capabilities) {
                if (channel.getState() != IChannel.STATE_OPEN) {
                    Protocol.invokeLater(done);
                    return;
                }
                channel.addChannelListener(new IChannel.IChannelListener() {
                    public void congestionLevel(int level) {
                    }

                    public void onChannelClosed(Throwable error) {
                        if (disposed) return;
                        channels.remove(channel);
                        if (channels.isEmpty()) {
                            id2bp.clear();
                        }
                    }

                    public void onChannelOpened() {
                    }
                });
                channels.put(channel, capabilities);
                Protocol.invokeLater(done);
            }
        });
    }

    public void setBreakpoints(List<BreakpointInfo> bpInfo, Predicate<BreakpointInfo> bpFilter, IChannel channel, CompletableFuture<BreakpointInfo[]> future) {
        assert !disposed;
        assert !Protocol.isDispatchThread();

        CompletableFuture<Map<String, BreakpointInfo>> filteredBpsFuture = new CompletableFuture<>();
        Protocol.invokeAndWait(() -> {
            final  IBreakpoints service = channel.getRemoteService(IBreakpoints.class);
            if (service == null) {
                filteredBpsFuture.complete(null);
            }
            if (bpInfo == null) {
                filteredBpsFuture.complete(null);
            }
            Map<String, BreakpointInfo> filteredBps = new HashMap<>();
            for (Map.Entry<String, BreakpointInfo> entry : id2bp.entrySet()) {
                if (bpFilter.test(entry.getValue())) {
                    filteredBps.put(entry.getKey(), entry.getValue());
                }
            }
            filteredBpsFuture.complete(filteredBps);
        });

        Map<String, BreakpointInfo> filteredBreakpoints = null;
        try {
            filteredBreakpoints = filteredBpsFuture.get();
        } catch (InterruptedException  | ExecutionException ex) {
            future.completeExceptionally(ex);
        }

        if (filteredBreakpoints == null) {
            future.complete(null);
            return;
        }
        for (BreakpointInfo bp : bpInfo) {
            String id = bp.getBreakpointId();
            if (id == null) continue;
            if (!filteredBreakpoints.containsKey(bp.getBreakpointId())) {
                addBreakpoint(bp);
            } else if (!filteredBreakpoints.get(id).equals(bp)) {
                changeBreakpoint(bp);
            }
        }
        List<String> newBreakpointIds = bpInfo.stream().map(bp -> bp.getBreakpointId()).collect(Collectors.toList());
        for (String id : filteredBreakpoints.keySet()) {
            if (newBreakpointIds.contains(id)) {
                continue;
            } else if (filteredBreakpoints.get(id).isForeign()) {
                continue;
            } else {
                removeBreakpoint(filteredBreakpoints.get(id));
            }
        }
        future.complete(bpInfo.toArray(new BreakpointInfo[0]));
    }

    public List<BreakpointInfo> getForeignBreakpoints(BreakpointInfo.Type type, IChannel channel) {
        // foreign breakpoints may be removed from the client. adding them back through events
        CompletableFuture<Map<String, BreakpointInfo>> filteredBpsFuture = new CompletableFuture<>();
        Protocol.invokeAndWait(() -> {
            final  IBreakpoints service = channel.getRemoteService(IBreakpoints.class);
            if (service == null) {
                filteredBpsFuture.complete(null);
            }
            Map<String, BreakpointInfo> filteredBps = new HashMap<>();
            for (Map.Entry<String, BreakpointInfo> entry : id2bp.entrySet()) {
                if (entry.getValue().getType() == type && entry.getValue().isForeign()) {
                    filteredBps.put(entry.getKey(), entry.getValue());
                }
            }
            filteredBpsFuture.complete(filteredBps);
        });

        Map<String, BreakpointInfo> foreignBreakpoints = null;
        try {
            foreignBreakpoints = filteredBpsFuture.get();
            if (foreignBreakpoints == null) {
                return null;
            }
            return foreignBreakpoints.entrySet().stream().map(Map.Entry::getValue).collect(Collectors.toList());
        } catch (InterruptedException  | ExecutionException ex) {
            return null;
        }
    }

/**
     * Download breakpoint info to a target.
     * This is supposed to be done once after a channel is opened.
     * @param channel - TCF communication channel.
     * @param done - client callback.
     * @throws IOException
     * @throws CoreException
     */
/* TODO: rm - this is broken into getBreakpointCapabilities and setBreakpoints

    @SuppressWarnings("unchecked")
    public void downloadBreakpoints(final IChannel channel, final Runnable done) throws IOException, CoreException {
        assert !disposed;
        assert Protocol.isDispatchThread();
        final IBreakpoints service = channel.getRemoteService(IBreakpoints.class);
        if (service == null) {
            Protocol.invokeLater(done);
            return;
        }
        service.getCapabilities(null, new IBreakpoints.DoneGetCapabilities() {
            public void doneGetCapabilities(IToken token, Exception error, Map<String,Object> capabilities) {
                if (channel.getState() != IChannel.STATE_OPEN) {
                    Protocol.invokeLater(done);
                    return;
                }
                if (channels.isEmpty()) {
                    bp_manager.addBreakpointListener(breakpoint_listener);
                    bp_manager.addBreakpointManagerListener(manager_listener);
                }
                channel.addChannelListener(new IChannel.IChannelListener() {
                    public void congestionLevel(int level) {
                    }
                    public void onChannelClosed(Throwable error) {
                        if (disposed) return;
                        channels.remove(channel);
                        if (channels.isEmpty()) {
                            bp_manager.removeBreakpointListener(breakpoint_listener);
                            bp_manager.removeBreakpointManagerListener(manager_listener);
                            id2bp.clear();
                        }
                    }
                    public void onChannelOpened() {
                    }
                });
                channels.put(channel, capabilities);
                IBreakpoint[] arr = bp_manager.getBreakpoints();
                if (arr != null && arr.length > 0) {
                    List<Map<String,Object>> bps = new ArrayList<Map<String,Object>>(arr.length);
                    for (int i = 0; i < arr.length; i++) {
                        try {
                            if (!isSupported(channel, arr[i])) continue;
                            String id = getBreakpointID(arr[i]);
                            if (id == null) continue;
                            if (!arr[i].isPersisted()) continue;
                            IMarker marker = arr[i].getMarker();
                            String file = getFilePath(marker.getResource());
                            bps.add(toBreakpointAttributes(channel, id, file, marker.getType(), marker.getAttributes()));
                            id2bp.put(id, arr[i]);
                        }
                        catch (Exception x) {
                            Activator.log("Cannot get breakpoint attributes", x);
                        }
                    }
                    if (!bps.isEmpty()) {
                        Map<String, Object>[] bp_arr = (Map<String,Object>[])bps.toArray(new Map[bps.size()]);
                        service.set(bp_arr, new IBreakpoints.DoneCommand() {
                            public void doneCommand(IToken token, Exception error) {
                                if (error == null) done.run();
                                else channel.terminate(error);
                            }
                        });
                        return;
                    }
                }
                Protocol.invokeLater(done);
            }
        });
    }
*/

    /**
     * Translate TCF breakpoint properties to Eclipse breakpoint marker attributes.
     * @param p - TCF breakpoint properties.
     * @return Eclipse marker attributes.
     */
/* TODO: fix

    public Map<String,Object> toMarkerAttributes(Map<String,Object> p) {
        assert !disposed;
        assert Protocol.isDispatchThread();
        Map<String,Object> client_data = (Map<String,Object>)p.get(IBreakpoints.PROP_CLIENT_DATA);
        if (client_data != null) {
            Map<String,Object> m = (Map<String,Object>)client_data.get(CDATA_MARKER);
            if (m != null) {
                m = new HashMap<String,Object>(m);
                m.put(ATTR_ID, p.get(IBreakpoints.PROP_ID));
                Boolean enabled = (Boolean)p.get(IBreakpoints.PROP_ENABLED);
                m.put(ATTR_ENABLED, enabled == null ? Boolean.FALSE : enabled);
                return m;
            }
        }
        Map<String,Object> m = new HashMap<String,Object>();
        for (Map.Entry<String,Object> e : p.entrySet()) {
            String key = e.getKey();
            Object val = e.getValue();
            if (key.equals(IBreakpoints.PROP_ENABLED)) continue;
            if (key.equals(IBreakpoints.PROP_FILE)) continue;
            if (key.equals(IBreakpoints.PROP_LINE)) continue;
            if (key.equals(IBreakpoints.PROP_COLUMN)) continue;
            if (key.equals(IBreakpoints.PROP_LOCATION)) continue;
            if (key.equals(IBreakpoints.PROP_ACCESS_MODE)) continue;
            if (key.equals(IBreakpoints.PROP_SIZE)) continue;
            if (key.equals(IBreakpoints.PROP_CONDITION)) continue;
            if (key.equals(IBreakpoints.PROP_EVENT_TYPE)) continue;
            if (key.equals(IBreakpoints.PROP_EVENT_ARGS)) continue;
            if (IBreakpoints.PROP_CONTEXT_IDS.equals(key) ||
                    IBreakpoints.PROP_CONTEXT_NAMES.equals(key) ||
                    IBreakpoints.PROP_STOP_GROUP.equals(key) ||
                    IBreakpoints.PROP_EXECUTABLE_PATHS.equals(key)) {
                if (val instanceof String[]) {
                    StringBuffer bf = new StringBuffer();
                    for (String s : (String[])val) {
                        if (bf.length() > 0) bf.append(',');
                        bf.append(s);
                    }
                    if (bf.length() == 0) continue;
                    val = bf.toString();
                }
                else if (val instanceof Collection) {
                    StringBuffer bf = new StringBuffer();
                    for (String s : (Collection<String>)val) {
                        if (bf.length() > 0) bf.append(',');
                        bf.append(s);
                    }
                    if (bf.length() == 0) continue;
                    val = bf.toString();
                }
            }
            if (!(val instanceof String) && !(val instanceof Boolean) && !(val instanceof Integer)) {
                // class of the value is not supported by Marker, convert to JSON string
                try {
                    val = "JSON:" + JSON.toJSON(val);
                }
                catch (IOException x) {
                    continue;
                }
            }
            m.put(ITCFConstants.ID_TCF_DEBUG_MODEL + '.' + key, val);
        }
        Boolean enabled = (Boolean)p.get(IBreakpoints.PROP_ENABLED);
        m.put(ATTR_ENABLED, enabled == null ? Boolean.FALSE : enabled);
        String location = (String)p.get(IBreakpoints.PROP_LOCATION);
        if (location != null && location.length() > 0) {
            int access_mode = IBreakpoints.ACCESSMODE_EXECUTE;
            Number access_mode_num = (Number)p.get(IBreakpoints.PROP_ACCESS_MODE);
            if (access_mode_num != null) access_mode = access_mode_num.intValue();
            if ((access_mode & IBreakpoints.ACCESSMODE_EXECUTE) != 0) {
                if (Character.isDigit(location.charAt(0))) {
                    m.put(ATTR_ADDRESS, location);
                }
                else {
                    m.put(ATTR_FUNCTION, location);
                }
            }
            else {
                m.put(ATTR_EXPRESSION, location.replaceFirst("^&\\((.+)\\)$", "$1"));
                m.put(ATTR_READ, (access_mode & IBreakpoints.ACCESSMODE_READ) != 0);
                m.put(ATTR_WRITE, (access_mode & IBreakpoints.ACCESSMODE_WRITE) != 0);
            }
            Number size_num = (Number)p.get(IBreakpoints.PROP_SIZE);
            if (size_num != null) m.put(ATTR_SIZE, size_num.toString());
        }
        m.put(IBreakpoint.REGISTERED, Boolean.TRUE);
        m.put(IBreakpoint.PERSISTED, Boolean.TRUE);
        m.put(IBreakpoint.ID, ITCFConstants.ID_TCF_DEBUG_MODEL);
        String file = (String)p.get(IBreakpoints.PROP_FILE);
        if (file != null && file.length() > 0) {
            m.put(ATTR_FILE, file);
        }
        Number line = (Number)p.get(IBreakpoints.PROP_LINE);
        if (line != null) {
            m.put(ATTR_LINE, Integer.valueOf(line.intValue()));
            Number column = (Number)p.get(IBreakpoints.PROP_COLUMN);
            if (column != null) {
                m.put(IMarker.CHAR_START, Integer.valueOf(column.intValue()));
                m.put(IMarker.CHAR_END, Integer.valueOf(column.intValue() + 1));
            }
        }
        String condition = (String)p.get(IBreakpoints.PROP_CONDITION);
        if (condition != null && condition.length() > 0) m.put(ATTR_CONDITION, condition);
        String event_type = (String)p.get(IBreakpoints.PROP_EVENT_TYPE);
        if (event_type != null && event_type.length() > 0) m.put(ATTR_EVENT_TYPE, event_type);
        String event_args = (String)p.get(IBreakpoints.PROP_EVENT_ARGS);
        if (event_args != null && event_args.length() > 0) m.put(ATTR_EVENT_ARGS, event_args);
        Number ignore_count = (Number)p.get(IBreakpoints.PROP_IGNORE_COUNT);
        if (ignore_count != null) m.put(ATTR_IGNORE_COUNT, ignore_count);
        Boolean temporary = (Boolean)p.get(IBreakpoints.PROP_TEMPORARY);
        if (temporary != null && temporary.booleanValue()) {
            Integer cdt_type = (Integer)m.get(ATTR_TYPE);
            cdt_type = cdt_type != null ? cdt_type : 0;
            cdt_type = cdt_type | ATTR_TYPE_TEMPORARY;
            m.put(ATTR_TYPE, cdt_type);
        }
        String type = (String)p.get(IBreakpoints.PROP_TYPE);
        if (type != null) {
            Integer cdt_type = (Integer)m.get(ATTR_TYPE);
            cdt_type = cdt_type != null ? cdt_type : 0;
            if (IBreakpoints.TYPE_HARDWARE.equals(type)) {
                cdt_type = cdt_type | ATTR_TYPE_HARDWARE;
            }
            else if (IBreakpoints.TYPE_SOFTWARE.equals(type)) {
                cdt_type = cdt_type | ATTR_TYPE_SOFTWARE;
            }
            m.put(ATTR_TYPE, cdt_type);
        }
        String msg = null;
        if (location != null) msg = location;
        else if (file != null && line != null) msg = file + ":" + line;
        else msg = (String)p.get(IBreakpoints.PROP_ID);
        m.put(ATTR_MESSAGE, "Breakpoint: " + msg);
        m.put(ATTR_TCF_STAMP, Boolean.TRUE.toString());

        return m;
    }
*/

/**
     * Translate Eclipse breakpoint marker attributes to TCF breakpoint properties.
     * @param channel - TCF communication channel.
     * @param bpInfo - breakpoint info.
     * @return TCF breakpoint properties.
     */
    public Map<String,Object> toBreakpointAttributes(IChannel channel, BreakpointInfo bpInfo) {
        assert !disposed;
        assert Protocol.isDispatchThread();
        Map<String,Object> m = new HashMap<String,Object>();
        Map<String,Object> capabilities = channels.get(channel);
        Map<String,Object> client_data = null;
        if (capabilities != null) {
            Object obj = capabilities.get(IBreakpoints.CAPABILITY_CLIENT_DATA);
            if (obj instanceof Boolean && ((Boolean)obj).booleanValue()) client_data = new HashMap<String,Object>();
        }
        m.put(IBreakpoints.PROP_ID, bpInfo.getBreakpointId());
        if (client_data != null) {
            m.put(IBreakpoints.PROP_CLIENT_DATA, client_data);
            client_data.put(CDATA_CLIENT_ID, "tcf-dap-1"); // FIXME: Unique client ID
//            client_data.put(CDATA_TYPE, type.toString());
        }
/*  TODO:fix Required for filtering breakpoints based on context

        for (Map.Entry<String,Object> e : p.entrySet()) {
            String key = e.getKey();
            Object val = e.getValue();
            if (key.equals(ATTR_STATUS)) continue;
            if (key.equals(ATTR_TCF_STAMP)) continue;
            if (key.startsWith(ITCFConstants.ID_TCF_DEBUG_MODEL)) {
                String tcf_key = key.substring(ITCFConstants.ID_TCF_DEBUG_MODEL.length() + 1);
                if (val instanceof String) {
                    String str = (String)val;
                    if (IBreakpoints.PROP_CONTEXT_IDS.equals(tcf_key)) {
                        if (str.length() == 0) continue;
                        val = filterContextIds(channel, str.split(",\\s*"));
                    }
                    else if (IBreakpoints.PROP_CONTEXT_QUERY.equals(tcf_key)) {
                        if (str.length() == 0) continue;
                    }
                    else if (IBreakpoints.PROP_CONTEXT_NAMES.equals(tcf_key) ||
                            IBreakpoints.PROP_STOP_GROUP.equals(tcf_key) ||
                            IBreakpoints.PROP_EXECUTABLE_PATHS.equals(tcf_key)) {
                        val = str.split(",\\s*");
                    }
                    else if (str.startsWith("JSON:")) {
                        try {
                            val = JSON.parseOne(str.substring(5).getBytes("UTF-8"));
                        }
                        catch (Exception x) {
                            // Not JSON, keep value as String
                        }
                    }
                }
                m.put(tcf_key, val);
            }
        }
*/
        m.put(IBreakpoints.PROP_ENABLED, bpInfo.isEnabled());
        String file = bpInfo.getFile();
        if (file != null && file.trim().length() > 0) {
            String name = file;
            boolean file_mapping = false;
            if (capabilities != null) {
                Object obj = capabilities.get(IBreakpoints.CAPABILITY_FILE_MAPPING);
                if (obj instanceof Boolean) file_mapping = ((Boolean)obj).booleanValue();
            }
            if (!file_mapping) {
                int i = file.lastIndexOf('/');
                int j = file.lastIndexOf('\\');
                if (i > j) name = file.substring(i + 1);
                else if (i < j) name = file.substring(j + 1);
            }
            m.put(IBreakpoints.PROP_FILE, name);
            Integer line = bpInfo.getLine();
            if (line != null && line >= 0) {
                m.put(IBreakpoints.PROP_LINE, Integer.valueOf(line.intValue()));
                Integer column = bpInfo.getColumn();
                if (column != null && column >= 0) m.put(IBreakpoints.PROP_COLUMN, column);
            }
        }

        if (bpInfo.getType() == BreakpointInfo.Type.DATA) {
/*  TODO:fix Add dataBreakpoint support here;

            String expr = (String)p.get(ATTR_EXPRESSION);
            if (expr != null && expr.length() != 0) {
                boolean writeAccess = Boolean.TRUE.equals(p.get(ATTR_WRITE));
                boolean readAccess = Boolean.TRUE.equals(p.get(ATTR_READ));
                int accessMode = 0;
                if (readAccess) accessMode |= IBreakpoints.ACCESSMODE_READ;
                if (writeAccess) accessMode |= IBreakpoints.ACCESSMODE_WRITE;
                m.put(IBreakpoints.PROP_ACCESS_MODE, Integer.valueOf(accessMode));
                Object range = p.get(ATTR_SIZE);
                if (range != null) {
                    int size = Integer.parseInt(range.toString());
                    if (size > 0) m.put(IBreakpoints.PROP_SIZE, size);
                }
                if (!Character.isDigit(expr.charAt(0))) {
                    expr = "&(" + expr + ')';
                }
                m.put(IBreakpoints.PROP_LOCATION, expr);
                if (m.get(IBreakpoints.PROP_FILE) != null && m.get(IBreakpoints.PROP_LINE) == null) {
                    m.put(IBreakpoints.PROP_LINE, Integer.valueOf(1));
                }
            }
*/
        }
        else if (bpInfo.getType() == BreakpointInfo.Type.FUNCTION) {
            String function = bpInfo.getFunction();
            if (function != null && function.length() != 0) {
                // Strip function parameters
                int paren = function.indexOf('(');
                if (paren > 0)
                    function = function.substring(0, paren);
                m.put(IBreakpoints.PROP_LOCATION, function);
            }
            if (!m.containsKey(IBreakpoints.PROP_SKIP_PROLOGUE) && capabilities != null) {
                Object obj = capabilities.get(IBreakpoints.CAPABILITY_SKIP_PROLOGUE);
                if (obj instanceof Boolean && ((Boolean)obj).booleanValue()) {
                    m.put(IBreakpoints.PROP_SKIP_PROLOGUE, true);
                }
            }
        }
/*
        else if (file == null) {
            String address = (String)p.get(ATTR_ADDRESS);
            if (address != null && address.length() > 0) m.put(IBreakpoints.PROP_LOCATION, address);
        }
*/
        String condition = bpInfo.getCondition();

        String logMessage = bpInfo.getLogMessage();
        if (logMessage != null && logMessage.length() > 0)
            if (condition != null && condition.length() > 0)
                condition = '(' + condition + ") && " + "$printf(" + logMessage + ')';
            else
                condition = "$printf(" + logMessage + ')';

        if (condition != null && condition.length() > 0) m.put(IBreakpoints.PROP_CONDITION, condition);
/*
        String event_type = (String)p.get(ATTR_EVENT_TYPE);
        if (event_type != null && event_type.length() > 0) m.put(IBreakpoints.PROP_EVENT_TYPE, event_type);
        String event_args = (String)p.get(ATTR_EVENT_ARGS);
        if (event_args != null && event_args.length() > 0) m.put(IBreakpoints.PROP_EVENT_ARGS, event_args);
*/
        int ignore_count = bpInfo.getIgnoreCount();
        if (ignore_count > 0) m.put(IBreakpoints.PROP_IGNORE_COUNT, ignore_count);
/*
        Integer cdt_type = (Integer)p.get(ATTR_TYPE);
        if (cdt_type != null) {
            if ((cdt_type.intValue() & ATTR_TYPE_TEMPORARY) != 0) {
                m.put(IBreakpoints.PROP_TEMPORARY, true);
            }
            if ((cdt_type.intValue() & ATTR_TYPE_HARDWARE) != 0) {
                m.put(IBreakpoints.PROP_TYPE, IBreakpoints.TYPE_HARDWARE);
            }
            else if ((cdt_type.intValue() & ATTR_TYPE_SOFTWARE) != 0) {
                m.put(IBreakpoints.PROP_TYPE, IBreakpoints.TYPE_SOFTWARE);
            }
            else  {
                m.put(IBreakpoints.PROP_TYPE, IBreakpoints.TYPE_AUTO);
            }
        }
*/
        m.put(IBreakpoints.PROP_TYPE, IBreakpoints.TYPE_AUTO);
        return m;
    }

    public BreakpointInfo createForeignBreakpointInfo(String foreignId, Map<String, Object> p, int localId) {
        assert !disposed;
        assert Protocol.isDispatchThread();
/*
        Map<String,Object> client_data = (Map<String,Object>)p.get(IBreakpoints.PROP_CLIENT_DATA);
        if (client_data != null) {
            Map<String,Object> m = (Map<String,Object>)client_data.get(CDATA_MARKER);
            if (m != null) {
                m = new HashMap<String,Object>(m);
                m.put(ATTR_ID, p.get(IBreakpoints.PROP_ID));
                Boolean enabled = (Boolean)p.get(IBreakpoints.PROP_ENABLED);
                m.put(ATTR_ENABLED, enabled == null ? Boolean.FALSE : enabled);
                return m;
            }
        }
*/
/*
        for (Map.Entry<String,Object> e : p.entrySet()) {
            String key = e.getKey();
            Object val = e.getValue();
            if (key.equals(IBreakpoints.PROP_ENABLED)) continue;
            if (key.equals(IBreakpoints.PROP_FILE)) continue;
            if (key.equals(IBreakpoints.PROP_LINE)) continue;
            if (key.equals(IBreakpoints.PROP_COLUMN)) continue;
            if (key.equals(IBreakpoints.PROP_LOCATION)) continue;
            if (key.equals(IBreakpoints.PROP_ACCESS_MODE)) continue;
            if (key.equals(IBreakpoints.PROP_SIZE)) continue;
            if (key.equals(IBreakpoints.PROP_CONDITION)) continue;
            if (key.equals(IBreakpoints.PROP_EVENT_TYPE)) continue;
            if (key.equals(IBreakpoints.PROP_EVENT_ARGS)) continue;
            if (IBreakpoints.PROP_CONTEXT_IDS.equals(key) ||
                    IBreakpoints.PROP_CONTEXT_NAMES.equals(key) ||
                    IBreakpoints.PROP_STOP_GROUP.equals(key) ||
                    IBreakpoints.PROP_EXECUTABLE_PATHS.equals(key)) {
                if (val instanceof String[]) {
                    StringBuffer bf = new StringBuffer();
                    for (String s : (String[])val) {
                        if (bf.length() > 0) bf.append(',');
                        bf.append(s);
                    }
                    if (bf.length() == 0) continue;
                    val = bf.toString();
                }
                else if (val instanceof Collection) {
                    StringBuffer bf = new StringBuffer();
                    for (String s : (Collection<String>)val) {
                        if (bf.length() > 0) bf.append(',');
                        bf.append(s);
                    }
                    if (bf.length() == 0) continue;
                    val = bf.toString();
                }
            }
            if (!(val instanceof String) && !(val instanceof Boolean) && !(val instanceof Integer)) {
                // class of the value is not supported by Marker, convert to JSON string
                try {
                    val = "JSON:" + JSON.toJSON(val);
                }
                catch (IOException x) {
                    continue;
                }
            }
            m.put(ITCFConstants.ID_TCF_DEBUG_MODEL + '.' + key, val);
        }
*/
        // id set to -1 to make sure it does not work without setting a local id
        BreakpointInfo bpInfo = BreakpointInfo.newForeignBreakpoint(foreignId, localId);
        Boolean enabled = (Boolean)p.get(IBreakpoints.PROP_ENABLED);
        bpInfo.setEnabled(enabled == null ? Boolean.FALSE : enabled);
        String location = (String)p.get(IBreakpoints.PROP_LOCATION);
        if (location != null && location.length() > 0) {
            int access_mode = IBreakpoints.ACCESSMODE_EXECUTE;
            Number access_mode_num = (Number)p.get(IBreakpoints.PROP_ACCESS_MODE);
            if (access_mode_num != null) access_mode = access_mode_num.intValue();
            if ((access_mode & IBreakpoints.ACCESSMODE_EXECUTE) != 0) {
                if (Character.isDigit(location.charAt(0))) {
//                    TODO: For data breakpoint
//                    bpInfo.setAddress(location);
                }
                else {
                    bpInfo.setType(BreakpointInfo.Type.FUNCTION);
                    bpInfo.setFunction(location);
                }
            }
//            else {
//                m.put(ATTR_EXPRESSION, location.replaceFirst("^&\\((.+)\\)$", "$1"));
//                m.put(ATTR_READ, (access_mode & IBreakpoints.ACCESSMODE_READ) != 0);
//                m.put(ATTR_WRITE, (access_mode & IBreakpoints.ACCESSMODE_WRITE) != 0);
//            }
//            Number size_num = (Number)p.get(IBreakpoints.PROP_SIZE);
//            if (size_num != null) m.put(ATTR_SIZE, size_num.toString());
        }
/*
        m.put(IBreakpoint.REGISTERED, Boolean.TRUE);
        m.put(IBreakpoint.PERSISTED, Boolean.TRUE);
        m.put(IBreakpoint.ID, ITCFConstants.ID_TCF_DEBUG_MODEL);
*/
        String file = (String)p.get(IBreakpoints.PROP_FILE);
        if (file != null && file.length() > 0) {
            bpInfo.setType(BreakpointInfo.Type.SOURCE);
            bpInfo.setFile(file);
        }
        Number line = (Number)p.get(IBreakpoints.PROP_LINE);
        if (line != null) {
            bpInfo.setLine(Integer.valueOf(line.intValue()));
            Number column = (Number)p.get(IBreakpoints.PROP_COLUMN);
            if (column != null) {
                bpInfo.setColumn(Integer.valueOf(column.intValue()));
            }
        }
        String condition = (String)p.get(IBreakpoints.PROP_CONDITION);
        if (condition != null && condition.length() > 0) bpInfo.setCondition(condition);
/*
        String event_type = (String)p.get(IBreakpoints.PROP_EVENT_TYPE);
        if (event_type != null && event_type.length() > 0) m.put(ATTR_EVENT_TYPE, event_type);
        String event_args = (String)p.get(IBreakpoints.PROP_EVENT_ARGS);
        if (event_args != null && event_args.length() > 0) m.put(ATTR_EVENT_ARGS, event_args);
*/
        Number ignore_count = (Number)p.get(IBreakpoints.PROP_IGNORE_COUNT);
        if (ignore_count != null) bpInfo.setIgnoreCount(ignore_count.intValue());
/*
        Boolean temporary = (Boolean)p.get(IBreakpoints.PROP_TEMPORARY);
        if (temporary != null && temporary.booleanValue()) {
            Integer cdt_type = (Integer)m.get(ATTR_TYPE);
            cdt_type = cdt_type != null ? cdt_type : 0;
            cdt_type = cdt_type | ATTR_TYPE_TEMPORARY;
            m.put(ATTR_TYPE, cdt_type);
        }
        String type = (String)p.get(IBreakpoints.PROP_TYPE);
        if (type != null) {
            Integer cdt_type = (Integer)m.get(ATTR_TYPE);
            cdt_type = cdt_type != null ? cdt_type : 0;
            if (IBreakpoints.TYPE_HARDWARE.equals(type)) {
                cdt_type = cdt_type | ATTR_TYPE_HARDWARE;
            }
            else if (IBreakpoints.TYPE_SOFTWARE.equals(type)) {
                cdt_type = cdt_type | ATTR_TYPE_SOFTWARE;
            }
            m.put(ATTR_TYPE, cdt_type);
        }
        String msg = null;
        if (location != null) msg = location;
        else if (file != null && line != null) msg = file + ":" + line;
        else msg = (String)p.get(IBreakpoints.PROP_ID);
        m.put(ATTR_MESSAGE, "Breakpoint: " + msg);
        m.put(ATTR_TCF_STAMP, Boolean.TRUE.toString());
*/
        return bpInfo;
    }

    /**
     * Filter given array of scope IDs of the form sessionId/contextId
     * to those applicable to the given channel.
     *//*
    private String[] filterContextIds(IChannel channel, String[] scopeIds) {
        String sessionId = getSessionId(channel);
        List<String> contextIds = new ArrayList<String>();
        for (String scopeId : scopeIds) {
            if (scopeId.length() == 0) continue;
            int slash = scopeId.indexOf('/');
            if (slash < 0) {
                contextIds.add(scopeId);
            }
            else if (sessionId != null && sessionId.equals(scopeId.substring(0, slash))) {
                contextIds.add(scopeId.substring(slash+1));
            }
        }
        return (String[]) contextIds.toArray(new String[contextIds.size()]);
    }

    *//**
     * @return launch config name for given channel or <code>null</code>
     *//*
    private String getSessionId(IChannel channel) {
        ILaunch[] launches = DebugPlugin.getDefault().getLaunchManager().getLaunches();
        for (ILaunch launch : launches) {
            if (launch instanceof TCFLaunch) {
                if (channel == ((TCFLaunch) launch).getChannel()) {
                    ILaunchConfiguration lc = launch.getLaunchConfiguration();
                    return lc != null ? lc.getName() : null;
                }
            }
        }
        return null;
    }
    */
}
